# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib.auth.models import AbstractBaseUser
from django.utils.translation import ugettext_lazy as _

from . import xmpp_backend


class XmppBackendUser(AbstractBaseUser):
    """An abstract base model using xmpp as backend for various functions.

    The model assumes that the username as returned by :py:func:`get_username` is the full JID of the user.
    """

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')
        abstract = True

    def exists(self):
        return xmpp_backend.user_exists(self.node, self.domain)

    def set_password(self, raw_password):
        """Calls :py:func:`~xmpp_backends.base.XmppBackendBase.set_password` for the user.

        If password is ``None``, calls :py:func:`~xmpp_backends.base.XmppBackendBase.set_unusable_password`.
        """
        if raw_password is None:
            self.set_unusable_password()
        else:
            xmpp_backend.set_password(self.node, self.domain, raw_password)

    def check_password(self, raw_password):
        """Calls :py:func:`~xmpp_backends.base.XmppBackendBase.check_password` for the user."""
        return xmpp_backend.check_password(self.node, self.domain, raw_password)

    def set_unusable_password(self):
        """Calls :py:func:`~xmpp_backends.base.XmppBackendBase.ban_account` for the user."""

        xmpp_backend.block_user(self.node, self.domain)

    def get_short_name(self):
        """An alias for ``node``."""

        return self.node

    @property
    def node(self):
        """The node-part of the username.

        Example::

            >>> u = XmppBackendUser(username='user@example.com')
            >>> u.node
            'user'
        """
        return self.get_username().split('@', 1)[0]

    @property
    def domain(self):
        """The domain-part of the username.

        Example::

            >>> u = XmppBackendUser(username='user@example.com')
            >>> u.domain
            'example.com'
        """
        return self.get_username().split('@', 1)[1]



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

CONNECTION_XMPP = 0
CONNECTION_HTTP_POLLING = 1
CONNECTION_HTTP_BINDING = 2
CONNECTION_WEBSOCKETS = 3
CONNECTION_UNKNOWN = 4



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import logging
from datetime import datetime
from datetime import timedelta
from subprocess import PIPE
from subprocess import Popen

import pytz
import six

from .base import BackendError
from .base import EjabberdBackendBase
from .base import NotSupportedError
from .base import UserExists
from .base import UserNotFound
from .base import UserSession

log = logging.getLogger(__name__)


class EjabberdctlBackend(EjabberdBackendBase):
    """This backend uses the ejabberdctl command line utility.

    This backend requires ejabberds ``mod_admin_extra`` to be installed.

    .. WARNING::

       This backend is not very secure because ejabberdctl gets any passwords in clear text via the
       commandline. The process list (and thus the passwords) can usually be viewed by anyone that has
       shell-access to your machine!

    :param    path: Optional path to the ``ejabberdctl`` script. The default is ``"/usr/sbin/ejabberdctl"``.
    :param version: A tuple describing the version used, e.g. ``(16, 12,)``. See :ref:`version parameter
        <ejabberd_version>` for a more detailed explanation.
    """

    def __init__(self, path='/usr/sbin/ejabberdctl', version=(17, 7, )):
        self.ejabberdctl = path
        self.version = version

        if version == (14, 7):
            log.warn('ejabberd 14.07 is really broken and many calls will not work!')
        if version < (13, 6):
            raise NotImplementedError('EjabberdRestBackend does not support ejabberd <= 16.01.')

    def get_version(self):
        return self.version

    def ex(self, *cmd):
        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        return p.returncode, stdout, stderr

    def ctl(self, *cmd):
        return self.ex(self.ejabberdctl, *cmd)

    def user_exists(self, username, domain):
        code, out, err = self.ctl('check_account', username, domain)
        if code == 0:
            return True
        elif code == 1:
            return False
        else:
            raise BackendError(code)  # TODO: 3 means nodedown.

    def user_sessions(self, username, domain):
        code, out, err = self.ctl('user_sessions_info', username, domain)
        version = self.get_version()
        sessions = set()
        out = out.decode('utf-8')  # bytes -> str in py3, str -> unicode in py2

        for line in out.splitlines():
            conn, ip, _p, prio, _n, uptime, status, resource, status_text = line.split('\t', 8)
            started = pytz.utc.localize(datetime.utcnow() - timedelta(int(uptime)))

            if prio == 'undefined':
                prio = None
            else:
                prio = int(prio)

            typ, encrypted, compressed = self.parse_connection_string(conn, version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=resource,
                priority=prio,
                ip_address=self.parse_ip_address(ip, version),
                uptime=started,
                status=status, status_text=status_text,
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))

        if len(sessions) == 0 and version == (14, 7):
            raise NotSupportedError("ejabberd 14.07 always returns an empty list.")

        return sessions

    def stop_user_session(self, username, domain, resource, reason=''):
        self.ctl('kick_session', username, domain, resource, reason)

    def create_user(self, username, domain, password, email=None):
        code, out, err = self.ctl('register', username, domain, password)

        if code == 0:
            try:
                self.set_last_activity(username, domain, status='Registered')
            except BackendError as e:
                log.error('Error setting last activity: %s', e)

            if email is not None:
                self.set_email(username, domain, email)
        elif code == 1:
            raise UserExists()
        else:
            raise BackendError(code)  # TODO: 3 means nodedown.

    def get_last_activity(self, username, domain):
        code, out, err = self.ctl('get_last', username, domain)
        version = self.get_version()

        if code != 0:
            raise BackendError(code)

        if six.PY3:
            out = out.decode('utf-8')

        if version < (17, 4):
            out = out.strip()
            if out == 'Online':
                return datetime.utcnow()
            elif out == 'Never':
                if self.user_exists(username, domain):
                    return None
                raise UserNotFound(username, domain)
            else:
                return datetime.strptime(out[:19], '%Y-%m-%d %H:%M:%S')
        else:
            timestamp, reason = out.strip().split('\t', 1)
            if reason == 'NOT FOUND':
                raise UserNotFound(username, domain)

            if len(timestamp) == 27:
                # NOTE: This format is encountered when the user is not found.
                fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
            else:
                fmt = '%Y-%m-%dT%H:%M:%SZ'

            return datetime.strptime(timestamp, fmt)

    def set_last_activity(self, username, domain, status='', timestamp=None):
        timestamp = str(self.datetime_to_timestamp(timestamp))
        code, out, err = self.ctl('set_last', username, domain, timestamp, status)

        version = self.get_version()
        if code == 1 and version >= (16, 1):
            # ejabberd returns status code 1 at least since 16.09
            return
        elif code != 0:
            if code == 1 and version == (14, 7):
                raise NotSupportedError("ejabberd 14.07 does not support setting last activity.")

            raise BackendError(code)

    def block_user(self, username, domain):
        # 14.07 is really broken here. ban_account fails, and we cannot set a random password
        # either because user_sessions_info doesn't work either and thus we can't stop existing
        # connections. And stopping existing connections doesn't work either.
        code, out, err = self.ctl('ban_account', username, domain, 'Blocked')
        if code != 0:
            version = self.get_version()
            if code == 1 and version == (14, 7):
                raise NotSupportedError("ejabberd 14.07 does not support banning accounts.")

            raise BackendError(code)

    def check_password(self, username, domain, password):
        code, out, err = self.ctl('check_password', username, domain, password)

        if code == 0:
            return True
        elif code == 1:
            return False
        else:
            raise BackendError(code)

    def set_password(self, username, domain, password):
        version = self.get_version()
        if version <= (16, 1, ) and not self.user_exists(username, domain):
            # 16.01 just creates the user upon change_password!
            # NOTE: This may also affect other versions < 16.09.
            raise UserNotFound(username, domain)

        code, out, err = self.ctl('change_password', username, domain, password)
        if six.PY3:
            out = out.decode('utf-8')

        if code == 1 and out == '{not_found,"unknown_user"}\n':
            raise UserNotFound(username, domain)
        elif code != 0:  # 0 is also returned if the user doesn't exist.
            raise BackendError(code)

    def set_unusable_password(self, username, domain):
        code, out, err = self.ctl('ban_account', username, domain, 'by xmpp-account')
        if code != 0:
            raise BackendError(code)

    def message_user(self, username, domain, subject, message):
        """Currently use send_message_chat and discard subject, because headline messages are not stored by
        mod_offline."""
        version = self.get_version()
        jid = '%s@%s' % (username, domain)
        if version <= (14, 7):
            # TODO: it's unclear when send_message was introduced
            command = 'send_message_chat'
            args = domain, '%s@%s' % (username, domain), message
        else:
            command = 'send_message'
            args = 'chat', domain, jid, subject, message

        code, out, err = self.ctl(command, *args)
        if code != 0:
            raise BackendError(code)

    def all_domains(self):
        code, out, err = self.ctl('registered_vhosts')
        if code != 0:
            raise BackendError(code)
        if six.PY3:
            out = out.decode('utf-8')

        return set(out.splitlines())

    def all_users(self, domain):
        code, out, err = self.ctl('registered_users', domain)
        if code != 0:
            raise BackendError(code)

        if six.PY3:
            out = out.decode('utf-8')

        return set(out.splitlines())

    def all_user_sessions(self):
        code, out, err = self.ctl('connected_users_info')
        if code != 0:
            raise BackendError(code)
        out = out.decode('utf-8')  # bytes -> str in py3, str -> unicode in py2
        version = self.get_version()
        sessions = set()

        for line in out.splitlines():
            jid, conn, ip, _p, prio, node, uptime = line.split('\t', 7)

            username, domain = jid.split('@', 1)
            domain, resource = domain.split('/', 1)
            if prio == 'nil':
                prio = None
            else:
                prio = int(prio)

            started = pytz.utc.localize(datetime.utcnow() - timedelta(int(uptime)))
            typ, encrypted, compressed = self.parse_connection_string(conn, version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=resource,
                priority=prio,
                ip_address=self.parse_ip_address(ip, version),
                uptime=started,
                status='',  # session['status'],
                status_text='',  # session['statustext'],
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))

        return sessions

    def remove_user(self, username, domain):
        code, out, err = self.ctl('unregister', username, domain)
        if code != 0:  # 0 is also returned if the user does not exist
            raise BackendError(code)

    def stats(self, stat, domain=None):
        if stat == 'registered_users':
            stat = 'registeredusers'
        elif stat == 'online_users':
            stat = 'onlineusers'
        else:
            raise ValueError("Unknown stat %s" % stat)

        if domain is None:
            code, out, err = self.ctl('stats', stat)
        else:
            code, out, err = self.ctl('stats_host', stat, domain)

        if code == 0:
            return int(out)
        else:
            raise BackendError(code)



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import socket
from datetime import datetime
from datetime import timedelta

import pytz
import six

from six.moves.http_client import BadStatusLine

from .base import BackendConnectionError
from .base import BackendError
from .base import EjabberdBackendBase
from .base import NotSupportedError
from .base import UserExists
from .base import UserNotFound
from .base import UserSession

if six.PY2:  # we have a special version for Python2
    from . import xmlrpclib
else:  # use the default for Python3 (god save us all!)
    from xmlrpc import client as xmlrpclib


log = logging.getLogger(__name__)


class EjabberdXMLRPCBackend(EjabberdBackendBase):
    """This backend uses the Ejabberd XMLRPC interface.

    In addition to an XMLRPC endpoint, this backend requires ``mod_admin_extra`` to be installed.

    .. WARNING::

        The backend does not handle UTF-8 characters correctly if you use ejabberd <= 14.07 and
        Python 3.

    **ejabberd configuration:** The ``xmlrpc`` module is included with ejabberd_ since version 13.12. If you
    use an earlier version, please get and run the module from the `ejabberd-contrib
    <https://github.com/processone/ejabberd-contrib>`_ repository. Configuring the interface is simple::

        listen:
            - ip: "127.0.0.1"
              port: 4560
              module: ejabberd_xmlrpc

    :param          uri: Directly passed to xmlrpclib, defaults to ``"http://127.0.0.1:4560"``.
    :param    transport: Directly passed to xmlrpclib.
    :param     encoding: Directly passed to xmlrpclib.
    :param      verbose: Directly passed to xmlrpclib.
    :param   allow_none: Directly passed to xmlrpclib.
    :param use_datetime: Directly passed to xmlrpclib.
    :param      context: Directly passed to xmlrpclib. Ignored in in Python3, where the parameter is still
        documented but no longer accepted by the ServerProxy constructor.
    :param         user: Username of the JID used for authentication.
    :param       server: Server of the JID used for authenticiation.
    :param     password: The password of the given JID.
    :param      version: A tuple describing the ejabberd version used, e.g. ``(16, 12,)``. See
        :ref:`version parameter <ejabberd_version>` for a more detailed explanation.
    """
    credentials = None

    def __init__(self, uri='http://127.0.0.1:4560', transport=None, encoding=None, verbose=0, allow_none=0,
                 use_datetime=0, context=None, user=None, server=None, password=None, version=(17, 7, )):
        super(EjabberdXMLRPCBackend, self).__init__()

        if version < (13, 6):
            raise NotImplementedError('EjabberdRestBackend does not support ejabberd <= 16.01.')
        kwargs = {
            'transport': transport,
            'encoding': encoding,
            'verbose': verbose,
            'allow_none': allow_none,
            'use_datetime': use_datetime,
        }
        if six.PY2 and version <= (14, 7, ):
            kwargs['utf8_encoding'] = 'php'
            kwargs['context'] = context

        self.client = xmlrpclib.ServerProxy(uri, **kwargs)
        if user is not None:
            self.credentials = {
                'user': user,
                'server': server,
                'password': password,
            }
        self.version = version

    def get_version(self):
        return self.version

    def rpc(self, cmd, **kwargs):
        """Generic helper function to call an RPC method."""

        func = getattr(self.client, cmd)
        try:
            if self.credentials is None:
                return func(kwargs)
            else:
                return func(self.credentials, kwargs)
        except socket.error as e:
            raise BackendConnectionError(e)
        except (xmlrpclib.ProtocolError, BadStatusLine) as e:
            log.error(e)
            raise BackendError("Error reaching backend.")

    def create_user(self, username, domain, password, email=None):
        result = self.rpc('register', user=username, host=domain, password=password)

        if result['res'] == 0:
            try:
                # we ignore errors here because not setting last activity is only a problem in edge-cases.
                self.set_last_activity(username, domain, status='Registered')
            except BackendError as e:
                log.error('Error setting last activity: %s', e)

            if email is not None:
                self.set_email(username, domain, email)
        elif result['res'] == 1:
            raise UserExists()
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def get_last_activity(self, username, domain):
        result = self.rpc('get_last', user=username, host=domain)
        version = self.get_version()

        if version < (17, 4):
            # ejabberd 17.04 introduced a change:
            #       https://github.com/processone/ejabberd/issues/1565
            activity = result['last_activity']
            if activity == 'Online':
                return datetime.utcnow()
            elif activity == 'Never':
                if self.user_exists(username, domain):
                    return None
                raise UserNotFound(username, domain)
            else:
                return datetime.strptime(activity[:19], '%Y-%m-%d %H:%M:%S')
        else:
            data = result['last_activity']
            if data[1]['status'] == 'NOT FOUND':
                raise UserNotFound(username, domain)

            timestamp = data[0]['timestamp']
            if len(timestamp) == 27:
                # NOTE: This format is encountered when the user is not found.
                fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
            else:
                fmt = '%Y-%m-%dT%H:%M:%SZ'

            return datetime.strptime(timestamp, fmt)

    def set_last_activity(self, username, domain, status='', timestamp=None):
        timestamp = self.datetime_to_timestamp(timestamp)
        self.rpc('set_last', user=username, host=domain, timestamp=timestamp, status=status)

    def user_exists(self, username, domain):
        result = self.rpc('check_account', user=username, host=domain)
        if result['res'] == 0:
            return True
        elif result['res'] == 1:
            return False
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def user_sessions(self, username, domain):
        version = self.get_version()
        result = self.rpc('user_sessions_info', user=username, host=domain)
        raw_sessions = result.get('sessions_info', [])
        sessions = set()
        for data in raw_sessions:
            # The data structure is a bit weird, its a list of one-element dicts.  We itemize each dict and
            # then flatten the resulting list
            session = [d.items() for d in data['session']]
            session = dict([item for sublist in session for item in sublist])

            started = pytz.utc.localize(datetime.utcnow() - timedelta(seconds=session['uptime']))
            typ, encrypted, compressed = self.parse_connection_string(session['connection'], version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=session['resource'],
                priority=session['priority'],
                ip_address=self.parse_ip_address(session['ip'], version),
                uptime=started,
                status=session['status'],
                status_text=session['statustext'],
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))
        return sessions

    def stop_user_session(self, username, domain, resource, reason=''):
        result = self.rpc('kick_session', user=username, host=domain, resource=resource,
                          reason=reason)
        return result

    def check_password(self, username, domain, password):
        result = self.rpc('check_password', user=username, host=domain, password=password)
        if result['res'] == 0:
            return True
        elif result['res'] == 1:
            return False
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def set_password(self, username, domain, password):
        version = self.get_version()
        if version <= (16, 1, ) and not self.user_exists(username, domain):
            # 16.01 just creates the user upon change_password!
            # NOTE: This may also affect other versions < 16.09.
            raise UserNotFound(username, domain)

        try:
            result = self.rpc('change_password', user=username, host=domain, newpass=password)
        except xmlrpclib.Fault as e:
            if e.faultCode == -118:
                raise UserNotFound(username, domain)
            raise BackendError('Unknown Error')

        if result['res'] == 0:
            return
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def block_user(self, username, domain):
        version = self.get_version()
        try:
            result = self.rpc('ban_account', user=username, host=domain, reason='Blocked.')
        except BackendError:
            if version == (14, 7):
                raise NotSupportedError('ejabberd 14.07 does not support getting all sessions via xmlrpc.')
            raise

        if result['res'] == 0:
            return
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def message_user(self, username, domain, subject, message):
        """Currently use send_message_chat and discard subject, because headline messages are not
        stored by mod_offline."""

        version = self.get_version()
        kwargs = {
            'body': message,
            'from': domain,
            'to': '%s@%s' % (username, domain),
        }

        if version <= (14, 7):
            # TODO: it's unclear when send_message was introduced
            command = 'send_message_chat'
        else:
            command = 'send_message'
            kwargs['subject'] = subject
            kwargs['type'] = 'normal'
        result = self.rpc(command, **kwargs)

        if result['res'] == 0:
            return
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def all_domains(self):
        return [d['vhost'] for d in self.rpc('registered_vhosts')['vhosts']]

    def all_users(self, domain):
        users = self.rpc('registered_users', host=domain)['users']
        return set([e['username'] for e in users])

    def all_user_sessions(self):
        version = self.get_version()
        try:
            result = self.rpc('connected_users_info')['connected_users_info']
        except BackendError:
            if version == (14, 7):
                raise NotSupportedError('ejabberd 14.07 does not support getting all sessions via xmlrpc.')
            raise
        sessions = set()
        for data in result:
            # The data structure is a bit weird, its a list of one-element dicts.  We itemize each dict and
            # then flatten the resulting list
            session = [d.items() for d in data['sessions']]
            session = dict([item for sublist in session for item in sublist])

            username, domain = session['jid'].split('@', 1)
            domain, resource = domain.split('/', 1)
            started = pytz.utc.localize(datetime.utcnow() - timedelta(seconds=session['uptime']))
            typ, encrypted, compressed = self.parse_connection_string(session['connection'], version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=resource,
                priority=session['priority'],
                ip_address=self.parse_ip_address(session['ip'], version),
                uptime=started,
                status='',  # session['status'],
                status_text='',  # session['statustext'],
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))
        return sessions

    def remove_user(self, username, domain):
        result = self.rpc('unregister', user=username, host=domain)
        if result['res'] == 0:
            return
        else:
            raise BackendError(result.get('text', 'Unknown Error'))

    def stats(self, stat, domain=None):
        if stat == 'registered_users':
            stat = 'registeredusers'
        elif stat == 'online_users':
            stat = 'onlineusers'
        else:
            raise ValueError("Unknown stat %s" % stat)

        if domain is None:
            result = self.rpc('stats', name=stat)
        else:
            result = self.rpc('stats_host', name=stat, host=domain)

        try:
            return result['stat']
        except KeyError:
            raise BackendError(result.get('text', 'Unknown Error'))



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import ipaddress
import logging
import time
from datetime import datetime

import pytz
import six

from .base import BackendError
from .base import UserExists
from .base import UserNotFound
from .base import UserSession
from .base import XmppBackendBase
from .constants import CONNECTION_XMPP

log = logging.getLogger(__name__)


class DummyBackend(XmppBackendBase):
    """A dummy backend for development using Djangos caching framework.

    By default, Djangos caching framework uses in-memory data structures, so every registration will be
    removed if you restart the development server.  You can configure a different cache (e.g. memcached), see
    `Django's cache framework <https://docs.djangoproject.com/en/dev/topics/cache/>`_ for details.

    :params domains: A list of domains to serve.
    """

    library = 'django.core.cache.cache'

    def __init__(self, domains):
        super(DummyBackend, self).__init__()
        self._domains = domains

    def user_exists(self, username, domain):
        if domain not in self._domains:
            return False

        user = '%s@%s' % (username, domain)
        return self.module.get(user) is not None

    def user_sessions(self, username, domain):
        user = '%s@%s' % (username, domain)
        return self.module.get(user).get('sessions', set())

    def start_user_session(self, username, domain, resource, **kwargs):
        """Method to add a user session for debugging.

        Accepted parameters are the same as to the constructor of :py:class:`~xmpp_backends.base.UserSession`.
        """

        kwargs.setdefault('uptime', pytz.utc.localize(datetime.utcnow()))
        kwargs.setdefault('priority', 0)
        kwargs.setdefault('status', 'online')
        kwargs.setdefault('status_text', '')
        kwargs.setdefault('connection_type', CONNECTION_XMPP)
        kwargs.setdefault('encrypted', True)
        kwargs.setdefault('compressed', False)
        kwargs.setdefault('ip_address', '127.0.0.1')
        if six.PY2 and isinstance(kwargs['ip_address'], str):
            # ipaddress constructor does not eat str in py2 :-/
            kwargs['ip_address'] = kwargs['ip_address'].decode('utf-8')
        if isinstance(kwargs['ip_address'], six.string_types):
            kwargs['ip_address'] = ipaddress.ip_address(kwargs['ip_address'])

        user = '%s@%s' % (username, domain)
        session = UserSession(self, username, domain, resource, **kwargs)

        data = self.module.get(user)
        if data is None:
            raise UserNotFound(username, domain, resource)

        data.setdefault('sessions', set())
        if isinstance(data['sessions'], list):
            # Cast old data to set
            data['sessions'] = set(data['sessions'])

        data['sessions'].add(session)
        self.module.set(user, data)

        all_sessions = self.module.get('all_sessions', set())
        all_sessions.add(session)
        self.module.set('all_sessions', all_sessions)

    def stop_user_session(self, username, domain, resource, reason=''):
        user = '%s@%s' % (username, domain)
        data = self.module.get(user)
        data['sessions'] = set([d for d in data.get('sessions', []) if d.resource != resource])
        self.module.set(user, data)

        all_sessions = self.module.get('all_sessions')
        all_sessions = set([s for s in all_sessions if s.jid != user])
        self.module.set('all_sessions', all_sessions)

    def create_user(self, username, domain, password, email=None):
        if domain not in self._domains:
            raise BackendError('Backend does not serve domain %s.' % domain)
        user = '%s@%s' % (username, domain)
        log.debug('Create user: %s (%s)', user, password)

        data = self.module.get(user)
        if data is None:
            data = {
                'pass': password,
                'last_status': (time.time(), 'Registered'),
                'sessions': set(),
            }
            if email is not None:
                data['email'] = email
            self.module.set(user, data)

            # maintain list of users in cache
            users = self.module.get('all_users', set())
            users.add(user)
            self.module.set('all_users', users)
        else:
            raise UserExists()

    def check_password(self, username, domain, password):
        user = '%s@%s' % (username, domain)
        log.debug('Check pass: %s -> %s', user, password)

        data = self.module.get(user)
        if data is None:
            return False
        else:
            return data['pass'] == password

    def check_email(self, username, domain, email):
        user = '%s@%s' % (username, domain)
        log.debug('Check email: %s --> %s', user, email)

        data = self.module.get(user)
        if data is None:
            return False
        else:
            return data['email'] == email

    def set_password(self, username, domain, password):
        user = '%s@%s' % (username, domain)
        log.debug('Set pass: %s -> %s', user, password)

        data = self.module.get(user)
        if data is None:
            raise UserNotFound(username, domain)
        else:
            data['pass'] = password
            self.module.set(user, data)

    def set_email(self, username, domain, email):
        user = '%s@%s' % (username, domain)
        log.debug('Set email: %s --> %s', user, email)

        data = self.module.get(user)
        if data is None:
            raise UserNotFound(username, domain)
        else:
            data['email'] = email
            self.module.set(user, data)

    def get_last_activity(self, username, domain):
        user = '%s@%s' % (username, domain)

        data = self.module.get(user)
        if data is None:
            raise UserNotFound(username, domain)
        else:
            return datetime.utcfromtimestamp(data['last_status'][0])

    def set_last_activity(self, username, domain, status='', timestamp=None):
        user = '%s@%s' % (username, domain)
        if timestamp is None:
            timestamp = time.time()
        else:
            timestamp = self.datetime_to_timestamp(timestamp)

        data = self.module.get(user)
        if data is None:
            pass  # NOTE: real APIs provide no error either :-/
        else:
            data['last_status'] = (timestamp, status)
            self.module.set(user, data)

    def block_user(self, username, domain):
        # overwritten so we pass tests
        self.set_password(username, domain, self.get_random_password())

    def all_domains(self):
        """Just returns the domains passed to the constructor."""

        return list(self._domains)

    def all_users(self, domain):
        return set([u.split('@')[0] for u in self.module.get('all_users', set())
                    if u.endswith('@%s' % domain)])

    def all_user_sessions(self):
        return self.module.get('all_sessions') or set()

    def remove_user(self, username, domain):
        user = '%s@%s' % (username, domain)
        log.debug('Remove: %s', user)

        users = self.module.get('all_users', set())
        users.remove(user)
        self.module.set('all_users', users)

        self.module.delete(user)

    def stats(self, stat, domain=None):
        """Always returns 0."""
        return 0



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
from datetime import datetime
from datetime import timedelta

import pytz
import requests

from .base import BackendConnectionError
from .base import BackendError
from .base import EjabberdBackendBase
from .base import UserExists
from .base import UserNotFound
from .base import UserSession

log = logging.getLogger(__name__)


class EjabberdRestBackend(EjabberdBackendBase):
    """This backend uses the Ejabberd REST interface.

    The module implements ejabberds still somewhat sparsely documented `REST API
    <https://docs.ejabberd.im/developer/ejabberd-api/>`_. It requires `requests
    <http://docs.python-requests.org/en/master/>`_.

    .. NOTE:: This backend requires ejabberd 16.02 or later.

    Our current (17.07) configuration looks as follows::

        # "api" is an access_rules list
        commands_admin_access: api
        commands:
          - add_commands:
            - change_password
            # ... other commands, see:
            # http://xmpp-backends.readthedocs.io/en/latest/xmpp_backends/backends.html#required-commands

        listen:
          # other ports here...
          - ip: "127.0.0.1"
            port: 5280
            module: ejabberd_http
            request_handlers:
              "/oauth": ejabberd_oauth
              "/api": mod_http_api

            # Requires ejabberd 17.04 or later, otherwise you should pass a version in the
            # constructor.
            custom_headers:
              "Ejabberd-Version": "@VERSION@"

            # Please use TLS if you're not on localhost! Help:
            #  http://xmpp-backends.readthedocs.io/en/latest/xmpp_backends/backends.html#ejabberd-tls-setup
            #tls: true
            #...

    :param       uri: The URI of the API.
    :param      user: User used in authenticating with the API.
    :param  password: Password used in authenticating with the API.
    :param   version: A tuple describing the version used, e.g. ``(16, 12,)``. See
        :ref:`version parameter <ejabberd_version>` for a more detailed explanation.
    :param \**kwargs: All keyword parameters are directly passed to the ``requests`` module. Please
        see the documentation there for possible parameters (e.g. SSL validation, etc.).
    """
    credentials = None

    def __init__(self, uri='http://127.0.0.1:5280/api/', user=None, password=None,
                 version=(17, 7, ), **kwargs):
        super(EjabberdRestBackend, self).__init__()

        if version <= (16, 1):
            raise NotImplementedError('EjabberdRestBackend does not support ejabberd <= 16.01.')

        if not uri.endswith('/'):
            uri += '/'

        self.uri = uri
        self.kwargs = kwargs
        self.headers = kwargs.pop('headers', {})
        self.headers.setdefault('X-Admin', 'true')
        self.version = version
        if user:
            self.kwargs['auth'] = (user, password)

    def get_version(self, response):
        if response.headers.get('Ejabberd-Version'):
            return response.headers.get('Ejabberd-Version').split('.')
        return self.version

    def post(self, cmd, allowed_status=None, **payload):
        if allowed_status is None:
            allowed_status = [200]

        uri = '%s%s' % (self.uri, cmd)
        try:
            response = requests.post(uri, json=payload, headers=self.headers, **self.kwargs)
        except requests.exceptions.RequestException as e:
            raise BackendConnectionError(e)

        if response.status_code not in allowed_status:
            raise BackendError('HTTP %s: %s' % (response.status_code, response.content))
        return response

    def create_user(self, username, domain, password, email=None):
        result = self.post('register', user=username, host=domain, password=password,
                           allowed_status=[200, 409])

        if result.status_code == 200:
            try:
                # we ignore errors here because not setting last activity is only a problem in
                # edge-cases.
                self.set_last_activity(username, domain, status='Registered')
            except BackendError as e:
                log.error('Error setting last activity: %s', e)

            if email is not None:
                self.set_email(username, domain, email)
        elif result.status_code == 409:
            raise UserExists()
        else:
            # NOTE: This really should never happen, only 200 and 409 don't raise an exception.
            raise BackendError(result.get('text', 'Unknown Error'))

    def get_last_activity(self, username, domain):
        response = self.post('get_last', user=username, host=domain)
        version = self.get_version(response)

        if version < (17, 4):
            result = response.json()['last_activity'].lower().strip()

            if result == 'never':
                if self.user_exists(username, domain):
                    # The output is the same if the user does not exist
                    return None
                raise UserNotFound(username, domain)
            elif result == 'online':
                return datetime.now()

            return datetime.strptime(result[:19], '%Y-%m-%d %H:%M:%S')

        # ejabberd 17.04 introduced a change:
        #       https://github.com/processone/ejabberd/issues/1565
        else:
            parsed = response.json()
            if parsed['status'] == 'NOT FOUND':
                raise UserNotFound(username, domain)

            timestamp = parsed['timestamp']
            if len(timestamp) == 27:
                # NOTE: This format is encountered when the user is not found.
                fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
            else:
                fmt = '%Y-%m-%dT%H:%M:%SZ'

            return datetime.strptime(parsed['timestamp'], fmt)

    def set_last_activity(self, username, domain, status='', timestamp=None):
        timestamp = self.datetime_to_timestamp(timestamp)
        self.post('set_last', user=username, host=domain, timestamp=timestamp, status=status)

    def user_exists(self, username, domain):
        response = self.post('check_account', user=username, host=domain)
        if response.content == b'0':
            return True
        elif response.content == b'1':
            return False
        else:
            raise BackendError('Unknown Error')

    def user_sessions(self, username, domain):
        response = self.post('user_sessions_info', user=username, host=domain)
        version = self.get_version(response)
        data = response.json()
        sessions = set()
        for d in data:
            started = pytz.utc.localize(datetime.utcnow() - timedelta(seconds=d['uptime']))
            typ, encrypted, compressed = self.parse_connection_string(d['connection'], version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=d['resource'],
                priority=d['priority'],
                ip_address=self.parse_ip_address(d['ip'], version),
                uptime=started,
                status=d['status'],
                status_text=d['statustext'],
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))
        return sessions

    def stop_user_session(self, username, domain, resource, reason=''):
        response = self.post('kick_session', user=username, host=domain, resource=resource,
                             reason=reason)
        return response

    def check_password(self, username, domain, password):
        response = self.post('check_password', user=username, host=domain, password=password)
        if response.content == b'0':
            return True
        elif response.content == b'1':
            return False
        else:
            raise BackendError('Unknown Error')

    def set_password(self, username, domain, password):
        response = self.post('change_password', user=username, host=domain, newpass=password,
                             allowed_status=[200, 404])
        if response.content == b'0':
            return
        elif response.status_code == 404:
            raise UserNotFound(username, domain)
        else:
            raise BackendError('Unknown Error')

    def block_user(self, username, domain):
        self.post('ban_account', user=username, host=domain, reason='Blocked.')

    def message_user(self, username, domain, subject, message):
        kwargs = {
            'body': message,
            'from': domain,
            'subject': subject,
            'to': '%s@%s' % (username, domain),
            'type': 'normal',
        }
        self.post('send_message', **kwargs)

    def all_domains(self):
        return self.post('registered_vhosts').json()

    def all_users(self, domain):
        return set(self.post('registered_users', host=domain).json())

    def all_user_sessions(self):
        # {'port': 49094, 'ip': '::1', 'connection': 'c2s', 'priority': 0, 'uptime': 3,
        #  'node': 'ejabberd@pallene', 'jid': 'example@example.com/3951214195792401555238'}
        response = self.post('connected_users_info')
        version = self.get_version(response)
        data = response.json()
        sessions = set()
        for d in data:
            started = pytz.utc.localize(datetime.utcnow() - timedelta(seconds=d['uptime']))
            username, domain = d['jid'].split('@', 1)
            domain, resource = domain.split('/', 1)

            typ, encrypted, compressed = self.parse_connection_string(d['connection'], version)
            sessions.add(UserSession(
                backend=self,
                username=username,
                domain=domain,
                resource=resource,
                priority=d['priority'],
                ip_address=self.parse_ip_address(d['ip'], version),
                uptime=started,
                status='',  # d['status'],
                status_text='',  # d['statustext'],
                connection_type=typ, encrypted=encrypted, compressed=compressed
            ))
        return sessions

    def remove_user(self, username, domain):
        response = self.post('unregister', user=username, host=domain)
        if response.content == b'""':
            return
        else:
            raise BackendError('Unknown Error')

    def stats(self, stat, domain=None):
        if stat == 'registered_users':
            stat = 'registeredusers'
        elif stat == 'online_users':
            stat = 'onlineusers'
        else:
            raise ValueError("Unknown stat %s" % stat)

        if domain is None:
            result = self.post('stats', name=stat).json()
        else:
            result = self.post('stats_host', name=stat, host=domain).json()

        try:
            return result['stat']
        except KeyError:
            raise BackendError(result.get('text', 'Unknown Error'))



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If
# not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, absolute_import

from importlib import import_module
from threading import local

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

DEFAULT_BACKEND_ALIAS = 'default'
__all__ = [
    'backend', 'DEFAULT_BACKEND_ALIAS', 'InvalidXmppBackendError',
]


class InvalidXmppBackendError(ImproperlyConfigured):
    pass


def _get_backend(backend, **kwargs):
    try:
        conf = settings.XMPP_BACKENDS[backend]
        params = conf.copy()
        params.update(kwargs)
        backend = params.pop('BACKEND')

        mod_path, cls_name = backend.rsplit('.', 1)
        mod = import_module(mod_path)
        backend_cls = getattr(mod, cls_name)
    except (KeyError, AttributeError, ImportError) as e:
        raise InvalidXmppBackendError(
            "Could not find backend '%s': %s" % (backend, e))
    backend = backend_cls(**params)
    return backend


class BackendHandler(object):
    """A Cache Handler to manage access to Cache instances.

    Ensures only one instance of each alias exists per thread.

    This implementation is an adaption of django.core.cache.BackendHandler.
    """
    def __init__(self):
        self._backends = local()

    def __getitem__(self, alias):
        try:
            return self._backends.backends[alias]
        except AttributeError:
            self._backends.backends = {}
        except KeyError:
            pass

        if alias not in settings.XMPP_BACKENDS:
            raise InvalidXmppBackendError(
                "Could not find config for '%s' in settings.XMPP_BACKENDS" % alias
            )

        backend = _get_backend(alias)
        self._backends.backends[alias] = backend
        return backend

    def all(self):
        return getattr(self._backends, 'backends', {}).values()


xmpp_backends = BackendHandler()


class DefaultBackendProxy(object):
    """Proxy access to the default backend object's attributes.

    This allows the `backend` object to be thread-safe using the `backends` API.

    This implementation is an adaption of django.core.cache.DefaultCacheProxy.
    """
    def __getattr__(self, name):
        return getattr(xmpp_backends[DEFAULT_BACKEND_ALIAS], name)

    def __setattr__(self, name, value):
        return setattr(xmpp_backends[DEFAULT_BACKEND_ALIAS], name, value)

    def __delattr__(self, name):
        return delattr(xmpp_backends[DEFAULT_BACKEND_ALIAS], name)

    def __contains__(self, key):
        return key in xmpp_backends[DEFAULT_BACKEND_ALIAS]

    def __eq__(self, other):
        return xmpp_backends[DEFAULT_BACKEND_ALIAS] == other

    def __ne__(self, other):
        return xmpp_backends[DEFAULT_BACKEND_ALIAS] != other


xmpp_backend = DefaultBackendProxy()



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib.auth import get_user_model

from . import xmpp_backend

User = get_user_model()


class XmppBackendBackend(object):
    """Authentication backend to authenticate against your XMPP server.

    To use this backend, simply configure the ``AUTHENTICATION_BACKENDS`` setting::

        AUTHENTICATION_BACKENDS = [
            'xmpp_backends.django.auth_backends.XmppBackendBackend',
        ]

    This authentication backend assumes that you store the full JID, including the domain, in the username
    field.
    """

    def authenticate(self, username, password):
        try:
            user = User.objects.get(**{User.USERNAME_FIELD: username, })

            node, domain = username.split('@', 1)
            if xmpp_backend.check_password(node, domain, password):
                return user
        except User.DoesNotExist:
            pass

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



#!/usr/bin/env python

import os
import sys

from setuptools import setup
from setuptools.command.test import test

PY2 = sys.version_info[0] == 2
install_requires = [
    'six>=1.10.0',
]

if PY2:
    install_requires.append('ipaddress>=1.0.18')


class django_test(test):
    def run(self):
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tests.django_settings')
        sys.path.insert(0, '.')
        import django
        django.setup()
        test.run(self)  # super() fails, LOL


setup(
    name='xmpp-backends',
    version='0.5.0',
    description='A set of classes with common interfaces to communicate with XMPP servers.',
    author='Mathias Ertl',
    author_email='mati@jabber.at',
    url='https://github.com/mathiasertl/xmpp-backends',
    packages=[
        'xmpp_backends',
        'xmpp_backends.django',
    ],
    cmdclass={
        'test': django_test,
    },
    license="GNU General Public License (GPL) v3",
    install_requires=install_requires,
    test_suite='tests',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Communications :: File Sharing",
    ],
    long_description="""A small python library providing a common interface for communicating with
the administration interfaces of various Jabber/XMPP servers."""
)



# -*- coding: utf-8 -*-
#
# This file is part of xmpp-backends (https://github.com/mathiasertl/xmpp-backends).
#
# xmpp-backends is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# xmpp-backends is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with xmpp-backends.  If not, see
# <http://www.gnu.org/licenses/>.

"""Common code for XMPP backends."""

from __future__ import unicode_literals

import ipaddress
import logging
import random
import string
import time
from datetime import datetime
from importlib import import_module

import pytz
import six

from .constants import CONNECTION_HTTP_BINDING
from .constants import CONNECTION_UNKNOWN
from .constants import CONNECTION_XMPP

log = logging.getLogger(__name__)


class BackendError(Exception):
    """All backend exceptions should be a subclass of this exception."""
    pass


class InvalidXmppBackendError(BackendError):
    """Raised when a module cannot be imported."""
    pass


class BackendConnectionError(BackendError):
    """Raised when the backend is unavailable."""
    pass


class NotSupportedError(BackendError):
    """Raised when a backend does not support a specific function.

    This error may be thrown only with specific versions, e.g. if it requires minimum version.
    """
    pass


class UserExists(BackendError):
    """Raised when a user already exists."""
    pass


class UserNotFound(BackendError):
    """Raised when a user is not found."""

    def __init__(self, node, domain, resource=None):
        self.node = node
        self.domain = domain
        self.resource = resource

    def __str__(self):
        s = '%s@%s' % (self.node, self.domain)
        if self.resource is not None:
            s += '/%s' % self.resource
        return s


@six.python_2_unicode_compatible
class UserSession(object):
    """An object describing a user session.

    :param         backend: The XMPP backend used for retrieving this session.
    :param        username: The username of the user.
    :type         username: str
    :param          domain: The domain of the user.
    :type           domain: str
    :param        resource: The resource of the user.
    :param        priority: The priority of this connection.
    :param      ip_address: The IP address of this connection.
    :param          uptime: A timestamp of when this connection came online.
    :param          status: The status message for this connection (e.g. "I am available.").
    :param connection_type: The type of connection.
    :param       encrypted: If this connection is encrypted. This may be ``None`` if the backend is not able
        decide if the connection is encrypted (e.g. if it is a HTTP bind connection).
    :param      compressed: If this connection uses XMPP stream compression. This is always ``None`` for
        connections where this is not applicable, e.g. Websocket connections.
    """
    def __init__(self, backend, username, domain, resource, priority, ip_address, uptime, status, status_text,
                 connection_type, encrypted, compressed):
        # make sure that data is in unicode
        if six.PY2 is True:
            if isinstance(username, six.binary_type):
                username = username.decode('utf-8')
            if isinstance(domain, six.binary_type):
                domain = domain.decode('utf-8')
            if isinstance(resource, six.binary_type):
                resource = resource.decode('utf-8')
            if isinstance(status_text, six.binary_type):
                status_text = status_text.decode('utf-8')

        self._backend = backend
        self.username = username
        self.domain = domain
        self.jid = '%s@%s' % (username, domain)
        self.resource = resource
        self.priority = priority
        self.ip_address = ip_address
        self.uptime = uptime
        self.status = status
        self.status_text = status_text
        self.connection_type = connection_type
        self.encrypted = encrypted
        self.compressed = compressed

    def __eq__(self, other):
        return isinstance(other, UserSession) and self.jid == other.jid and self.resource == other.resource

    def __hash__(self):
        return hash((self.jid, self.resource))

    def __str__(self):
        return '%s@%s/%s' % (self.username, self.domain, self.resource)

    def __repr__(self):
        val = '<UserSession: %s@%s/%s>' % (self.username, self.domain, self.resource)
        if six.PY2 is True:
            return val.encode('utf-8')
        return val


class XmppBackendBase(object):
    """Base class for all XMPP backends."""

    library = None
    """Import-party of any third-party library you need.

    Set this attribute to an import path and you will be able to access the module as ``self.module``. This
    way you don't have to do a module-level import, which would mean that everyone has to have that library
    installed, even if they're not using your backend.
    """
    _module = None

    @property
    def module(self):
        """The module specified by the ``library`` attribute."""

        if self._module is None:
            if self.library is None:
                raise ValueError(
                    "Backend '%s' doesn't specify a library attribute" % self.__class__)

            try:
                if '.' in self.library:
                    mod_path, cls_name = self.library.rsplit('.', 1)
                    mod = import_module(mod_path)
                    self._module = getattr(mod, cls_name)
                else:
                    self._module = import_module(self.library)
            except (AttributeError, ImportError):
                raise ValueError("Couldn't load %s backend library" % cls_name)

        return self._module

    def datetime_to_timestamp(self, dt):
        """Helper function to convert a datetime object to a timestamp.

        If datetime instance ``dt`` is naive, it is assumed that it is in UTC.

        In Python 3, this just calls ``datetime.timestamp()``, in Python 2, it substracts any timezone offset
        and returns the difference since 1970-01-01 00:00:00.

        Note that the function always returns an int, even in Python 3.

        >>> XmppBackendBase().datetime_to_timestamp(datetime(2017, 9, 17, 19, 59))
        1505678340
        >>> XmppBackendBase().datetime_to_timestamp(datetime(1984, 11, 6, 13, 21))
        468595260

        :param dt: The datetime object to convert. If ``None``, returns the current time.
        :type  dt: datetime
        :return: The seconds in UTC.
        :rtype: int
        """
        if dt is None:
            return int(time.time())

        if six.PY3:
            if not dt.tzinfo:
                dt = pytz.utc.localize(dt)
            return int(dt.timestamp())
        else:
            if dt.tzinfo:
                dt = dt.replace(tzinfo=None) - dt.utcoffset()
            return int((dt - datetime(1970, 1, 1)).total_seconds())

    def get_random_password(self, length=32, chars=None):
        """Helper function that gets a random password.

        :param length: The length of the random password.
        :type  length: int
        :param  chars: A string with characters to choose from. Defaults to all ASCII letters and digits.
        :type   chars: str
        """
        if chars is None:
            chars = string.ascii_letters + string.digits
        return ''.join(random.choice(chars) for x in range(length))

    def user_exists(self, username, domain):
        """Verify that the given user exists.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :return: ``True`` if the user exists, ``False`` if not.
        :rtype: bool
        """

        raise NotImplementedError

    def user_sessions(self, username, domain):
        """Get a list of all current sessions for the given user.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :return: A list :py:class:`~xmpp_backends.base.UserSession` describing the user sessions.
        :rtype: list of :py:class:`~xmpp_backends.base.UserSession`
        """

        raise NotImplementedError

    def stop_user_session(self, username, domain, resource, reason=''):
        """Stop a specific user session, identified by its resource.

        A resource uniquely identifies a connection by a specific client.

        :param   username: The username of the user.
        :type    username: str
        :param     domain: The domain of the user.
        :type      domain: str
        :param   resource: The resource of the connection
        :type    resource: str
        """

        raise NotImplementedError

    def create_user(self, username, domain, password, email=None):
        """Create a new user.

        :param username: The username of the new user.
        :type  username: str
        :param   domain: The domain of the new user.
        :type    domain: str
        :param password: The password of the new user.
        :param    email: The email address provided by the user.
        """
        raise NotImplementedError

    def create_reservation(self, username, domain, email=None):
        """Reserve a new account.

        This method is called when a user account should be reserved, meaning that the account can no longer
        be registered by anybody else but the user cannot yet log in either. This is useful if e.g. an email
        confirmation is still pending.

        The default implementation calls :py:func:`~xmpp_backends.base.XmppBackendBase.create_user` with a
        random password.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :param    email: The email address provided by the user. Note that at this point it is not confirmed.
            You are free to ignore this parameter.
        """
        password = self.get_random_password()
        self.create(username=username, domain=domain, password=password, email=email)

    def confirm_reservation(self, username, domain, password, email=None):
        """Confirm a reservation for a username.

        The default implementation just calls :py:func:`~xmpp_backends.base.XmppBackendBase.set_password` and
        optionally :py:func:`~xmpp_backends.base.XmppBackendBase.set_email`.
        """
        self.set_password(username=username, domain=domain, password=password)
        if email is not None:
            self.set_email(username=username, domain=domain, email=email)

    def check_password(self, username, domain, password):
        """Check the password of a user.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :param password: The password to check.
        :type  password: str
        :return: ``True`` if the password is correct, ``False`` if not.
        :rtype: bool
        """
        raise NotImplementedError

    def set_password(self, username, domain, password):
        """Set the password of a user.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :param password: The password to set.
        :type  password: str
        """
        raise NotImplementedError

    def get_last_activity(self, username, domain):
        """Get the last activity of the user.

        The datetime object returned should be a naive datetime object representing the time in UTC.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :return: A naive datetime object in UTC representing the last activity.
        :rtype: datetime
        """
        raise NotImplementedError

    def set_last_activity(self, username, domain, status='', timestamp=None):
        """Set the last activity of the user.

        .. NOTE::

           If your backend requires a Unix timestamp (seconds since 1970-01-01), you can use the
           :py:func:`~xmpp_backends.base.XmppBackendBase.datetime_to_timestamp` convenience function to
           convert it to an integer.

        :param  username: The username of the user.
        :type   username: str
        :param    domain: The domain of the user.
        :type     domain: str
        :param    status: The status text.
        :type     status: str
        :param timestamp: A datetime object representing the last activity. If the object is not
            timezone-aware, assume UTC. If ``timestamp`` is ``None``, assume the current date and time.
        :type  timestamp: datetime
        """
        raise NotImplementedError

    def block_user(self, username, domain):
        """Block the specified user.

        The default implementation calls :py:func:`~xmpp_backends.base.XmppBackendBase.set_password` with a
        random password.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        """
        self.set_password(username, domain, self.get_random_password())

    def set_email(self, username, domain, email):
        """Set the email address of a user."""
        raise NotImplementedError

    def check_email(self, username, domain, email):
        """Check the email address of a user.

        **Note:** Most backends don't implement this feature.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        """
        raise NotImplementedError

    def expire_reservation(self, username, domain):
        """Expire a username reservation.

        This method is called when a reservation expires. The default implementation just calls
        :py:func:`~xmpp_backends.base.XmppBackendBase.remove_user`. This is fine if you do not override
        :py:func:`~xmpp_backends.base.XmppBackendBase.create_reservation`.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        """
        self.remove_user(username, domain)

    def message_user(self, username, domain, subject, message):
        """Send a message to the given user.

        :param username: The username of the user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        :param  subject: The subject of the message.
        :param  message: The content of the message.
        """
        pass

    def all_users(self, domain):
        """Get all users for a given domain.

        :param   domain: The domain of interest.
        :type    domain: str
        :return: A set of all users. The usernames do not include the domain, so ``user@example.com`` will
            just be ``"user"``.
        :rtype: set of str
        """
        raise NotImplementedError

    def all_domains(self):
        """List of all domains used by this backend.

        :return: List of all domains served by this backend.
        :rtype: list of str
        """
        raise NotImplementedError

    def all_user_sessions(self):
        """List all current user sessions.

        :param domain: Optionally only return sessions for the given domain.
        :return: A list :py:class:`~xmpp_backends.base.UserSession` for all sessions.
        :rtype: list of :py:class:`~xmpp_backends.base.UserSession`
        """

        raise NotImplementedError

    def remove_user(self, username, domain):
        """Remove a user.

        This method is called when the user explicitly wants to remove her/his account.

        :param username: The username of the new user.
        :type  username: str
        :param   domain: The domain of the user.
        :type    domain: str
        """
        raise NotImplementedError

    def stats(self, stat, domain=None):
        """Get statistical value about the XMPP server.

        Minimal statistics that should be supported is ``"registered_users"`` and ``"online_users"``. The
        specific backend might support additional stats.

        :param stat: The value of the statistic.
        :type  stat: str
        :param domain: Limit statistic to the given domain. If not listed, give statistics
                       about all users.
        :type  domain: str
        :return: The current value of the requested statistic.
        :rtype: int
        """
        raise NotImplementedError


class EjabberdBackendBase(XmppBackendBase):
    """Base class for ejabberd related backends.

    This class overwrites a few methods common to all ejabberd backends.
    """
    def has_usable_password(self, username, domain):
        """Always return ``True``.

        In ejabberd there is no such thing as a "banned" account or an unusable password. Even ejabberd's
        ``ban_account`` command only sets a random password that the user could theoretically guess.
        """

        return True

    def set_email(self, username, domain, email):
        """Not yet implemented."""
        pass

    def check_email(self, username, domain, email):
        """Not yet implemented."""
        pass

    def parse_connection_string(self, connection, version):
        """Parse string as returned by the ``connected_users_info`` or ``user_sessions_info`` API calls.

        >>> EjabberdBackendBase().parse_connection_string('c2s_tls', (17, 8, ))
        (0, True, False)
        >>> EjabberdBackendBase().parse_connection_string('c2s_compressed_tls', (17, 8, ))
        (0, True, True)
        >>> EjabberdBackendBase().parse_connection_string('http_bind', (17, 8, ))
        (2, None, None)

        :param connection: The connection string as returned by the ejabberd APIs.
        :type  connection: str
        :param    version: The ejabberd version as tuple, currently unused.
        :type     version: tuple
        :return: A tuple representing the conntion type, if it is encrypted and if it uses XMPP stream
            compression.
        :rtype: tuple
        """
        # TODO: Websockets, HTTP Polling
        if connection == 'c2s_tls':
            return CONNECTION_XMPP, True, False
        elif connection == 'c2s_compressed_tls':
            return CONNECTION_XMPP, True, True
        elif connection == 'http_bind':
            return CONNECTION_HTTP_BINDING, None, None
        elif connection == 'c2s':
            return CONNECTION_XMPP, False, False
        log.warn('Could not parse connection string "%s"', connection)
        return CONNECTION_UNKNOWN, True, True

    def parse_ip_address(self, ip_address, version):
        """Parse an address as returned by the ``connected_users_info`` or ``user_sessions_info`` API calls.

        Example::

            >>> EjabberdBackendBase().parse_ip_address('192.168.0.1', (17, 8, ))  # doctest: +FORCE_TEXT
            IPv4Address('192.168.0.1')
            >>> EjabberdBackendBase().parse_ip_address('::1', (17, 8, ))  # doctest: +FORCE_TEXT
            IPv6Address('::1')

        :param ip_address: An IP address.
        :type  ip_address: str
        :param    version: The ejabberd version as tuple, currently unused.
        :type     version: tuple
        :return: The parsed IP address.
        :rtype: `ipaddress.IPv6Address` or `ipaddress.IPv4Address`.
        """
        if ip_address.startswith('::FFFF:'):
            ip_address = ip_address[7:]
        if six.PY2 and isinstance(ip_address, str):
            # ipaddress constructor does not eat str in py2 :-/
            ip_address = ip_address.decode('utf-8')

        return ipaddress.ip_address(ip_address)
